@@ -48,7 +48,7 @@ def build():
 
     # Parse the supplied command-line arguments
     try:
-        config = BuildConfiguration(parser, sys.argv[1:])
+        config = BuildConfiguration(parser, sys.argv[1:], logger)
     except RuntimeError as e:
         logger.error("Error: {}".format(e))
         sys.exit(1)
@@ -88,14 +88,11 @@ def build():
     # Create an auto-deleting temporary directory to hold our build context
     with tempfile.TemporaryDirectory() as tempDir:
 
-        # Copy our Dockerfiles to the temporary directory
         contextOrig = join(os.path.dirname(os.path.abspath(__file__)), "dockerfiles")
-        contextRoot = join(tempDir, "dockerfiles")
-        shutil.copytree(contextOrig, contextRoot)
 
         # Create the builder instance to build the Docker images
         builder = ImageBuilder(
-            contextRoot,
+            join(tempDir, "dockerfiles"),
             config.containerPlatform,
             logger,
             config.rebuild,
@@ -105,11 +102,12 @@ def build():
             config.combine,
         )
 
-        # Resolve our main set of tags for the generated images
-        mainTags = [
-            "{}{}-{}".format(config.release, config.suffix, config.prereqsTag),
-            config.release + config.suffix,
-        ]
+        # Resolve our main set of tags for the generated images; this is used only for Source and downstream
+        if config.buildTargets["source"]:
+            mainTags = [
+                "{}{}-{}".format(config.release, config.suffix, config.prereqsTag),
+                config.release + config.suffix,
+            ]
 
         # Print the command-line invocation that triggered this build, masking any supplied passwords
         args = [
@@ -128,10 +126,11 @@ def build():
         )
         if config.custom == True:
             logger.info("Custom name:   " + config.release, False)
-        else:
+        elif config.release is not None:
             logger.info("Release:       " + config.release, False)
-        logger.info("Repository:    " + config.repository, False)
-        logger.info("Branch/tag:    " + config.branch + "\n", False)
+        if config.repository is not None:
+            logger.info("Repository:    " + config.repository, False)
+            logger.info("Branch/tag:    " + config.branch + "\n", False)
 
         # Determine if we are using a custom version for ue4cli or conan-ue4cli
         if config.ue4cliVersion is not None or config.conanUe4cliVersion is not None:
@@ -235,13 +234,13 @@ def build():
                 )
                 sys.exit(1)
             elif newer_check is None:
-                logger.info(
+                logger.warning(
                     "Warning: unable to determine whether host system is new enough to use specified base tag"
                 )
 
             # Ensure the Docker daemon is configured correctly
             requiredLimit = WindowsUtils.requiredSizeLimit()
-            if DockerUtils.maxsize() < requiredLimit:
+            if DockerUtils.maxsize() < requiredLimit and config.buildTargets["source"]:
                 logger.error("SETUP REQUIRED:")
                 logger.error(
                     "The max image size for Windows containers must be set to at least {}GB.".format(
@@ -262,23 +261,28 @@ def build():
                 sys.exit(1)
 
         elif config.containerPlatform == "linux":
-
-            # Determine if we are building CUDA-enabled container images
-            capabilities = (
-                "CUDA {} + OpenGL".format(config.cuda)
-                if config.cuda is not None
-                else "OpenGL"
-            )
             logger.info("LINUX CONTAINER SETTINGS", False)
             logger.info(
-                "Building GPU-enabled images compatible with NVIDIA Docker ({} support).\n".format(
-                    capabilities
-                ),
+                "Base OS image: {}\n".format(config.baseImage),
                 False,
             )
 
         # Report which Engine components are being excluded (if any)
         logger.info("GENERAL SETTINGS", False)
+        logger.info(
+            "Build targets: {}".format(
+                " ".join(
+                    sorted(
+                        [
+                            target
+                            for target, enabled in config.buildTargets.items()
+                            if enabled
+                        ]
+                    )
+                )
+            ),
+            False,
+        )
         logger.info(
             "Changelist override: {}".format(
                 config.changelist
@@ -314,7 +318,10 @@ def build():
             username = ""
             password = ""
 
-        elif builder.willBuild("ue4-source", mainTags) == False:
+        elif (
+            not config.buildTargets["source"]
+            or builder.willBuild("ue4-source", mainTags) == False
+        ):
 
             # Don't bother prompting the user for any credentials if we're not building the ue4-source image
             logger.info(
@@ -338,11 +345,8 @@ def build():
         if config.args.monitor == True:
             resourceMonitor.start()
 
-        # Start the HTTP credential endpoint as a child process and wait for it to start
+        # Prep for endpoint cleanup, if necessary
         endpoint = None
-        if config.opts.get("credential_mode", "endpoint") == "endpoint":
-            endpoint = CredentialEndpoint(username, password)
-            endpoint.start()
 
         try:
 
@@ -363,91 +367,109 @@ def build():
                 "NAMESPACE={}".format(GlobalConfiguration.getTagNamespace()),
             ]
 
-            # Compute the build options for the UE4 build prerequisites image
-            # (This is the only image that does not use any user-supplied tag suffix, since the tag always reflects any customisations)
-            prereqsArgs = ["--build-arg", "BASEIMAGE=" + config.baseImage]
-            if config.containerPlatform == "windows":
-                prereqsArgs = prereqsArgs + [
+            # Build the UE4 build prerequisites image
+            if config.buildTargets["build-prerequisites"]:
+                # Compute the build options for the UE4 build prerequisites image
+                # (This is the only image that does not use any user-supplied tag suffix, since the tag always reflects any customisations)
+                prereqsArgs = ["--build-arg", "BASEIMAGE=" + config.baseImage]
+                if config.containerPlatform == "windows":
+                    prereqsArgs = prereqsArgs + [
+                        "--build-arg",
+                        "DLLSRCIMAGE=" + config.dllSrcImage,
+                        "--build-arg",
+                        "VISUAL_STUDIO_BUILD_NUMBER=" + config.visualStudioBuildNumber,
+                    ]
+
+                custom_prerequisites_dockerfile = config.args.prerequisites_dockerfile
+                if custom_prerequisites_dockerfile is not None:
+                    builder.build_builtin_image(
+                        "ue4-base-build-prerequisites",
+                        [config.prereqsTag],
+                        commonArgs + config.platformArgs + prereqsArgs,
+                        builtin_name="ue4-build-prerequisites",
+                    )
+                    builtImages.append("ue4-base-build-prerequisites")
+                else:
+                    builder.build_builtin_image(
+                        "ue4-build-prerequisites",
+                        [config.prereqsTag],
+                        commonArgs + config.platformArgs + prereqsArgs,
+                    )
+
+                prereqConsumerArgs = [
                     "--build-arg",
-                    "DLLSRCIMAGE=" + config.dllSrcImage,
-                    "--build-arg",
-                    "VISUAL_STUDIO_BUILD_NUMBER=" + config.visualStudioBuildNumber,
+                    "PREREQS_TAG={}".format(config.prereqsTag),
                 ]
 
-            # Build the UE4 build prerequisites image
-            builder.build(
-                "ue4-build-prerequisites",
-                [config.prereqsTag],
-                commonArgs + config.platformArgs + prereqsArgs,
-            )
-            builtImages.append("ue4-build-prerequisites")
+                if custom_prerequisites_dockerfile is not None:
+                    builder.build(
+                        "ue4-build-prerequisites",
+                        [config.prereqsTag],
+                        commonArgs + config.platformArgs + prereqConsumerArgs,
+                        dockerfile_template=custom_prerequisites_dockerfile,
+                        context_dir=os.path.dirname(custom_prerequisites_dockerfile),
+                    )
 
-            # If we're using build secrets then pass the Git username and password to the UE4 source image as secrets
-            secrets = {}
-            if config.opts.get("credential_mode", "endpoint") == "secrets":
-                secrets = {"username": username, "password": password}
+                builtImages.append("ue4-build-prerequisites")
+            else:
+                logger.info("Skipping ue4-build-prerequisities image build.")
 
             # Build the UE4 source image
-            prereqConsumerArgs = [
-                "--build-arg",
-                "PREREQS_TAG={}".format(config.prereqsTag),
-            ]
-            credentialArgs = [] if len(secrets) > 0 else endpoint.args()
-            ue4SourceArgs = prereqConsumerArgs + [
-                "--build-arg",
-                "GIT_REPO={}".format(config.repository),
-                "--build-arg",
-                "GIT_BRANCH={}".format(config.branch),
-                "--build-arg",
-                "VERBOSE_OUTPUT={}".format("1" if config.verbose == True else "0"),
-            ]
-            builder.build(
-                "ue4-source",
-                mainTags,
-                commonArgs + config.platformArgs + ue4SourceArgs + credentialArgs,
-                secrets,
-            )
-            builtImages.append("ue4-source")
+            if config.buildTargets["source"]:
+                # Start the HTTP credential endpoint as a child process and wait for it to start
+                if config.opts.get("credential_mode", "endpoint") == "endpoint":
+                    endpoint = CredentialEndpoint(username, password)
+                    endpoint.start()
 
-            # Build the UE4 Engine source build image, unless requested otherwise by the user
-            ue4BuildArgs = prereqConsumerArgs + [
-                "--build-arg",
-                "TAG={}".format(mainTags[1]),
-            ]
-            if config.noEngine == False:
-                builder.build(
-                    "ue4-engine",
+                # If we're using build secrets then pass the Git username and password to the UE4 source image as secrets
+                secrets = {}
+                if config.opts.get("credential_mode", "endpoint") == "secrets":
+                    secrets = {"username": username, "password": password}
+                credentialArgs = [] if len(secrets) > 0 else endpoint.args()
+
+                ue4SourceArgs = prereqConsumerArgs + [
+                    "--build-arg",
+                    "GIT_REPO={}".format(config.repository),
+                    "--build-arg",
+                    "GIT_BRANCH={}".format(config.branch),
+                    "--build-arg",
+                    "VERBOSE_OUTPUT={}".format("1" if config.verbose == True else "0"),
+                ]
+                builder.build_builtin_image(
+                    "ue4-source",
                     mainTags,
-                    commonArgs + config.platformArgs + ue4BuildArgs,
+                    commonArgs + config.platformArgs + ue4SourceArgs + credentialArgs,
+                    secrets=secrets,
                 )
-                builtImages.append("ue4-engine")
+                builtImages.append("ue4-source")
             else:
-                logger.info(
-                    "User specified `--no-engine`, skipping ue4-engine image build."
-                )
+                logger.info("Skipping ue4-source image build.")
+
+            if config.buildTargets["minimal"]:
+                ue4BuildArgs = prereqConsumerArgs + [
+                    "--build-arg",
+                    "TAG={}".format(mainTags[1]),
+                ]
 
             # Build the minimal UE4 CI image, unless requested otherwise by the user
-            minimalArgs = (
-                ["--build-arg", "CHANGELIST={}".format(config.changelist)]
-                if config.changelist is not None
-                else []
-            )
-            buildUe4Minimal = config.noMinimal == False
-            if buildUe4Minimal == True:
-                builder.build(
+            if config.buildTargets["minimal"]:
+                minimalArgs = (
+                    ["--build-arg", "CHANGELIST={}".format(config.changelist)]
+                    if config.changelist is not None
+                    else []
+                )
+
+                builder.build_builtin_image(
                     "ue4-minimal",
                     mainTags,
                     commonArgs + config.platformArgs + ue4BuildArgs + minimalArgs,
                 )
                 builtImages.append("ue4-minimal")
             else:
-                logger.info(
-                    "User specified `--no-minimal`, skipping ue4-minimal image build."
-                )
+                logger.info("Skipping ue4-minimal image build.")
 
             # Build the full UE4 CI image, unless requested otherwise by the user
-            buildUe4Full = buildUe4Minimal == True and config.noFull == False
-            if buildUe4Full == True:
+            if config.buildTargets["full"]:
 
                 # If custom version strings were specified for ue4cli and/or conan-ue4cli, use them
                 infrastructureFlags = []
@@ -467,7 +489,7 @@ def build():
                     )
 
                 # Build the image
-                builder.build(
+                builder.build_builtin_image(
                     "ue4-full",
                     mainTags,
                     commonArgs
@@ -477,9 +499,7 @@ def build():
                 )
                 builtImages.append("ue4-full")
             else:
-                logger.info(
-                    "Not building ue4-minimal or user specified `--no-full`, skipping ue4-full image build."
-                )
+                logger.info("Skipping ue4-full image build.")
 
             # If we are generating Dockerfiles then include information about the options used to generate them
             if config.layoutDir is not None:
