@@ -14,17 +14,15 @@ DEFAULT_GIT_REPO = "https://github.com/EpicGames/UnrealEngine.git"
 
 # The base images for Linux containers
 LINUX_BASE_IMAGES = {
-    "opengl": "nvidia/opengl:1.0-glvnd-devel-ubuntu18.04",
-    "cudagl": {
-        "9.2": "nvidia/cudagl:9.2-devel-ubuntu18.04",
-        "10.0": "nvidia/cudagl:10.0-devel-ubuntu18.04",
-        "10.1": "nvidia/cudagl:10.1-devel-ubuntu18.04",
-        "10.2": "nvidia/cudagl:10.2-devel-ubuntu18.04",
-    },
+    "opengl": "nvidia/opengl:1.0-glvnd-devel-{ubuntu}",
+    "cudagl": "nvidia/cudagl:{cuda}-devel-{ubuntu}",
 }
 
+# The default ubuntu base to use
+DEFAULT_LINUX_VERSION = "ubuntu18.04"
+
 # The default CUDA version to use when `--cuda` is specified without a value
-DEFAULT_CUDA_VERSION = "9.2"
+DEFAULT_CUDA_VERSION = "11.4"
 
 # The default memory limit (in GB) under Windows
 DEFAULT_MEMORY_LIMIT = 10.0
@@ -39,16 +37,19 @@ UNREAL_ENGINE_RELEASE_CHANGELISTS = {
     "4.25.0": 13144385,
     "4.26.0": 14830424,
     "4.27.0": 17155196,
+    "5.0.0": 19505902,
 }
 
 
 class VisualStudio(object):
     VS2017 = "2017"
     VS2019 = "2019"
+    VS2022 = "2022"
 
     BuildNumbers = {
         VS2017: "15",
         VS2019: "16",
+        VS2022: "17",
     }
 
     MinSupportedUnreal = {
@@ -57,7 +58,8 @@ class VisualStudio(object):
         #
         # Unreal Engine 4.25 is the first that works with .NET SDK 4.7+
         # See https://github.com/EpicGames/UnrealEngine/commit/5256eedbdef30212ab69fdf4c09e898098959683
-        VS2019: semver.VersionInfo(4, 25)
+        VS2019: semver.VersionInfo(4, 25),
+        VS2022: semver.VersionInfo(4, 27),
     }
 
 
@@ -95,6 +97,12 @@ class BuildConfiguration(object):
         """
         parser.add_argument(
             "release",
+            nargs="?",  # aka "required = False", but that doesn't work in positionals
+            help='UE4 release to build, in semver format (e.g. 4.20.0) or "custom" for a custom repo and branch (deprecated, use --ue-version instead)',
+        )
+        parser.add_argument(
+            "--ue-version",
+            default=None,
             help='UE4 release to build, in semver format (e.g. 4.20.0) or "custom" for a custom repo and branch',
         )
         parser.add_argument(
@@ -112,20 +120,24 @@ class BuildConfiguration(object):
             action="store_true",
             help="Print `docker build` commands instead of running them",
         )
-        parser.add_argument(
-            "--no-engine", action="store_true", help="Don't build the ue4-engine image"
-        )
         parser.add_argument(
             "--no-minimal",
             action="store_true",
-            help="Don't build the ue4-minimal image",
+            help="Don't build the ue4-minimal image (deprecated, use --target instead)",
         )
         parser.add_argument(
-            "--no-full", action="store_true", help="Don't build the ue4-full image"
+            "--no-full",
+            action="store_true",
+            help="Don't build the ue4-full image (deprecated, use --target instead)",
         )
         parser.add_argument(
             "--no-cache", action="store_true", help="Disable Docker build cache"
         )
+        parser.add_argument(
+            "--target",
+            action="append",
+            help="Add a target to the build list. Valid targets are `build-prerequisites`, `source`, `engine`, `minimal`, `full`, and `all`. May be specified multiple times or comma-separated. Defaults to `all`.",
+        )
         parser.add_argument(
             "--random-memory",
             action="store_true",
@@ -187,8 +199,9 @@ class BuildConfiguration(object):
         )
         parser.add_argument(
             "-basetag",
-            default=None,
-            help="Windows Server Core base image tag to use for Windows containers (default is the host OS version)",
+            default=None if platform.system() == "Windows" else DEFAULT_LINUX_VERSION,
+            help="Operating system base image tag to use. For Linux this is the version of Ubuntu (default is ubuntu18.04). "
+            "For Windows this is the Windows Server Core base image tag (default is the host OS version)",
         )
         parser.add_argument(
             "-suffix", default="", help="Add a suffix to the tags of the built images"
@@ -246,8 +259,13 @@ class BuildConfiguration(object):
             default=None,
             help="Set a specific changelist number in the Unreal Engine's Build.version file",
         )
+        parser.add_argument(
+            "--prerequisites-dockerfile",
+            default=None,
+            help="Specifies path to custom ue4-build-prerequisites dockerfile",
+        )
 
-    def __init__(self, parser, argv):
+    def __init__(self, parser, argv, logger):
         """
         Creates a new build configuration based on the supplied arguments object
         """
@@ -259,60 +277,167 @@ class BuildConfiguration(object):
         self.args = parser.parse_args(argv)
         self.changelist = self.args.changelist
 
-        # Determine if we are building a custom version of UE4 rather than an official release
-        self.args.release = self.args.release.lower()
-        if self.args.release == "custom" or self.args.release.startswith("custom:"):
+        # Figure out what targets we have; this is needed to find out if we need --ue-version.
+        using_target_specifier_old = self.args.no_minimal or self.args.no_full
+        using_target_specifier_new = self.args.target is not None
 
-            # Both a custom repository and a custom branch/tag must be specified
-            if self.args.repo is None or self.args.branch is None:
-                raise RuntimeError(
-                    "both a repository and branch/tag must be specified when building a custom version of the Engine"
-                )
-
-            # Use the specified repository and branch/tag
-            customName = (
-                self.args.release.split(":", 2)[1].strip()
-                if ":" in self.args.release
-                else ""
+        # If we specified nothing, it's the same as specifying `all`
+        if not using_target_specifier_old and not using_target_specifier_new:
+            self.args.target = ["all"]
+        elif using_target_specifier_old and not using_target_specifier_new:
+            # Convert these to the new style
+            logger.warning(
+                "Using deprecated `--no-*` target specifiers; recommend changing to `--target`",
+                False,
             )
-            self.release = customName if len(customName) > 0 else "custom"
-            self.repository = self.args.repo
-            self.branch = self.args.branch
-            self.custom = True
 
-        else:
+            # no-minimal implies no-full
+            if self.args.no_minimal:
+                self.args.no_full = True
 
-            # Validate the specified version string
-            try:
-                ue4Version = semver.parse(self.args.release)
-                if (
-                    ue4Version["major"] not in [4, 5]
-                    or ue4Version["prerelease"] != None
-                ):
-                    raise Exception()
-                self.release = semver.format_version(
-                    ue4Version["major"], ue4Version["minor"], ue4Version["patch"]
-                )
-            except:
+            # Change into target descriptors
+            self.args.target = []
+
+            if not self.args.no_full:
+                self.args.target += ["full"]
+
+            if not self.args.no_minimal:
+                self.args.target += ["minimal"]
+
+            # disabling these was never supported
+            self.args.target += ["source"]
+            self.args.target += ["build-prerequisites"]
+
+        elif using_target_specifier_new and not using_target_specifier_old:
+            # these can be token-delimited, so let's just split them apart and then remerge them into one list
+            split = [item.split(",") for item in self.args.target]
+            self.args.target = [item for sublist in split for item in sublist]
+
+        elif using_target_specifier_old and using_target_specifier_new:
+            # uhoh
+            raise RuntimeError(
+                "specified both `--target` and the old `--no-*` options; please use only `--target`!"
+            )
+
+        # Now that we have our options in `self.args.target`, evaluate our dependencies
+        # In a theoretical ideal world this should be code-driven; if you find yourself adding a lot more code to this, consider a redesign!
+        active_targets = set(self.args.target)
+
+        # build-prereq -> source -> engine
+        # build-prereq -> source -> minimal -> full
+
+        # We initialize these with all the options, with the intent that you should be accessing them directly and not checking for existence
+        # This is to avoid typos giving false-negatives; KeyError is reliable and tells you what you did wrong
+        self.buildTargets = {
+            "build-prerequisites": False,
+            "source": False,
+            "minimal": False,
+            "full": False,
+        }
+
+        for target in active_targets:
+            if target != "all" and target not in self.buildTargets:
+                valid_options = sorted(self.buildTargets.keys())
                 raise RuntimeError(
-                    'invalid Unreal Engine release number "{}", full semver format required (e.g. "4.20.0")'.format(
-                        self.args.release
-                    )
+                    f"unknown build target '{target}', valid options are: all {' '.join(valid_options)}"
                 )
 
-            # Use the default repository and the release tag for the specified version
-            self.repository = DEFAULT_GIT_REPO
-            self.branch = "{}-release".format(self.release)
-            self.custom = False
+        if "full" in active_targets or "all" in active_targets:
+            self.buildTargets["full"] = True
+            active_targets.add("minimal")
 
-            # If the user specified a .0 release of the Unreal Engine and did not specify a changelist override then
-            # use the official changelist number for that release to ensure consistency with Epic Games Launcher builds
-            # (This is necessary because .0 releases do not include a `CompatibleChangelist` value in Build.version)
-            if (
-                self.changelist is None
-                and self.release in UNREAL_ENGINE_RELEASE_CHANGELISTS
-            ):
-                self.changelist = UNREAL_ENGINE_RELEASE_CHANGELISTS[self.release]
+        if "minimal" in active_targets or "all" in active_targets:
+            self.buildTargets["minimal"] = True
+            active_targets.add("source")
+
+        if "source" in active_targets or "all" in active_targets:
+            self.buildTargets["source"] = True
+            active_targets.add("build-prerequisites")
+
+        if "build-prerequisites" in active_targets or "all" in active_targets:
+            self.buildTargets["build-prerequisites"] = True
+
+        if not self.buildTargets["build-prerequisites"]:
+            raise RuntimeError(
+                "we're not building anything; this shouldn't even be possible, but is definitely not useful"
+            )
+
+        # See if the user specified both the old positional version option and the new ue-version option
+        if self.args.release is not None and self.args.ue_version is not None:
+            raise RuntimeError(
+                "specified both `--ue-version` and the old positional version option; please use only `--ue-version`!"
+            )
+
+        # For the sake of a simpler pull request, we use self.args.release as the canonical place for this data.
+        # If support for the old positional version option is removed, this should be fixed.
+        if self.args.ue_version is not None:
+            self.args.release = self.args.ue_version
+
+        # We care about the version number only if we're building source
+        if self.buildTargets["source"]:
+            if self.args.release is None:
+                raise RuntimeError("missing `--ue-version` when building source")
+
+            # Determine if we are building a custom version of UE4 rather than an official release
+            self.args.release = self.args.release.lower()
+            if self.args.release == "custom" or self.args.release.startswith("custom:"):
+
+                # Both a custom repository and a custom branch/tag must be specified
+                if self.args.repo is None or self.args.branch is None:
+                    raise RuntimeError(
+                        "both a repository and branch/tag must be specified when building a custom version of the Engine"
+                    )
+
+                # Use the specified repository and branch/tag
+                customName = (
+                    self.args.release.split(":", 2)[1].strip()
+                    if ":" in self.args.release
+                    else ""
+                )
+                self.release = customName if len(customName) > 0 else "custom"
+                self.repository = self.args.repo
+                self.branch = self.args.branch
+                self.custom = True
+
+            else:
+
+                # Validate the specified version string
+                try:
+                    ue4Version = semver.parse(self.args.release)
+                    if (
+                        ue4Version["major"] not in [4, 5]
+                        or ue4Version["prerelease"] != None
+                    ):
+                        raise Exception()
+                    self.release = semver.format_version(
+                        ue4Version["major"], ue4Version["minor"], ue4Version["patch"]
+                    )
+                except:
+                    raise RuntimeError(
+                        'invalid Unreal Engine release number "{}", full semver format required (e.g. "4.20.0")'.format(
+                            self.args.release
+                        )
+                    )
+
+                # Use the default repository and the release tag for the specified version
+                self.repository = DEFAULT_GIT_REPO
+                self.branch = "{}-release".format(self.release)
+                self.custom = False
+
+                # If the user specified a .0 release of the Unreal Engine and did not specify a changelist override then
+                # use the official changelist number for that release to ensure consistency with Epic Games Launcher builds
+                # (This is necessary because .0 releases do not include a `CompatibleChangelist` value in Build.version)
+                if (
+                    self.changelist is None
+                    and self.release in UNREAL_ENGINE_RELEASE_CHANGELISTS
+                ):
+                    self.changelist = UNREAL_ENGINE_RELEASE_CHANGELISTS[self.release]
+        else:
+            # defaults needed by other parts of the codebase
+            self.custom = False
+            self.release = None
+            self.repository = None
+            self.branch = None
 
         # Store our common configuration settings
         self.containerPlatform = (
@@ -322,9 +447,6 @@ class BuildConfiguration(object):
         )
         self.dryRun = self.args.dry_run
         self.rebuild = self.args.rebuild
-        self.noEngine = self.args.no_engine
-        self.noMinimal = self.args.no_minimal
-        self.noFull = self.args.no_full
         self.suffix = self.args.suffix
         self.platformArgs = ["--no-cache"] if self.args.no_cache == True else []
         self.excludedComponents = set(self.args.exclude)
@@ -368,27 +490,29 @@ class BuildConfiguration(object):
                 "the `-layout` flag must be used when specifying the `--combine` flag"
             )
 
-        # Verify that the value for `source_mode` is valid if specified
-        validSourceModes = ["git", "copy"]
-        if self.opts.get("source_mode", "git") not in validSourceModes:
-            raise RuntimeError(
-                "invalid value specified for the `source_mode` option, valid values are {}".format(
-                    validSourceModes
+        # We care about source_mode and credential_mode only if we're building source
+        if self.buildTargets["source"]:
+            # Verify that the value for `source_mode` is valid if specified
+            validSourceModes = ["git", "copy"]
+            if self.opts.get("source_mode", "git") not in validSourceModes:
+                raise RuntimeError(
+                    "invalid value specified for the `source_mode` option, valid values are {}".format(
+                        validSourceModes
+                    )
                 )
-            )
 
-        # Verify that the value for `credential_mode` is valid if specified
-        validCredentialModes = (
-            ["endpoint", "secrets"]
-            if self.containerPlatform == "linux"
-            else ["endpoint"]
-        )
-        if self.opts.get("credential_mode", "endpoint") not in validCredentialModes:
-            raise RuntimeError(
-                "invalid value specified for the `credential_mode` option, valid values are {} when building {} containers".format(
-                    validCredentialModes, self.containerPlatform.title()
-                )
+            # Verify that the value for `credential_mode` is valid if specified
+            validCredentialModes = (
+                ["endpoint", "secrets"]
+                if self.containerPlatform == "linux"
+                else ["endpoint"]
             )
+            if self.opts.get("credential_mode", "endpoint") not in validCredentialModes:
+                raise RuntimeError(
+                    "invalid value specified for the `credential_mode` option, valid values are {} when building {} containers".format(
+                        validCredentialModes, self.containerPlatform.title()
+                    )
+                )
 
         # Generate Jinja context values for keeping or excluding components
         self.opts["excluded_components"] = {
@@ -397,6 +521,33 @@ class BuildConfiguration(object):
             "templates": ExcludedComponent.Templates in self.excludedComponents,
         }
 
+        # Warn user that they are in danger of Docker 20GB COPY bug
+        # Unfortunately, we don't have a cheap way to check whether user environment is affected
+        # See https://github.com/adamrehn/ue4-docker/issues/99
+        if self.containerPlatform == "windows":
+            warn20GiB = False
+            if ExcludedComponent.Debug not in self.excludedComponents:
+                logger.warning("Warning: You didn't pass --exclude debug", False)
+                warn20GiB = True
+            if (
+                self.release
+                and not self.custom
+                and semver.VersionInfo.parse(self.release) >= semver.VersionInfo(5, 0)
+            ):
+                logger.warning("Warning: You're building Unreal Engine 5", False)
+                warn20GiB = True
+
+            if warn20GiB:
+                logger.warning("Warning: You might hit Docker 20GiB COPY bug", False)
+                logger.warning(
+                    "Warning: Make sure that `ue4-docker diagnostics 20gig` passes",
+                    False,
+                )
+                logger.warning(
+                    "Warning: See https://github.com/adamrehn/ue4-docker/issues/99#issuecomment-1079702817 for details and workarounds",
+                    False,
+                )
+
         # If we're building Windows containers, generate our Windows-specific configuration settings
         if self.containerPlatform == "windows":
             self._generateWindowsConfig()
@@ -422,7 +573,7 @@ class BuildConfiguration(object):
     def _generateWindowsConfig(self):
         self.visualStudio = self.args.visual_studio
 
-        if not self.custom:
+        if self.release is not None and not self.custom:
             # Check whether specified Unreal Engine release is compatible with specified Visual Studio
             vsMinSupportedUnreal = VisualStudio.MinSupportedUnreal.get(
                 self.visualStudio, None
@@ -469,7 +620,7 @@ class BuildConfiguration(object):
             self.isolation = self.args.isolation
         else:
 
-            # If we are able to use process isolation mode then use it, otherwise fallback to the Docker daemon's default isolation mode
+            # If we are able to use process isolation mode then use it, otherwise use Hyper-V isolation mode
             differentKernels = self.basetag != self.hostBasetag
             dockerSupportsProcess = parse_version(
                 DockerUtils.version()["Version"]
@@ -477,7 +628,7 @@ class BuildConfiguration(object):
             if not differentKernels and dockerSupportsProcess:
                 self.isolation = "process"
             else:
-                self.isolation = DockerUtils.info()["Isolation"]
+                self.isolation = "hyperv"
 
         # Set the isolation mode Docker flag
         self.platformArgs.append("--isolation=" + self.isolation)
@@ -520,19 +671,19 @@ class BuildConfiguration(object):
 
             # Verify that the specified CUDA version is valid
             self.cuda = self.args.cuda if self.args.cuda != "" else DEFAULT_CUDA_VERSION
-            if self.cuda not in LINUX_BASE_IMAGES["cudagl"]:
-                raise RuntimeError(
-                    'unsupported CUDA version "{}", supported versions are: {}'.format(
-                        self.cuda, ", ".join([v for v in LINUX_BASE_IMAGES["cudagl"]])
-                    )
-                )
-
             # Use the appropriate base image for the specified CUDA version
-            self.baseImage = LINUX_BASE_IMAGES["cudagl"][self.cuda]
-            self.prereqsTag = "cudagl{}".format(self.cuda)
+            self.baseImage = LINUX_BASE_IMAGES["cudagl"]
+            self.prereqsTag = "cudagl{cuda}-{ubuntu}"
         else:
             self.baseImage = LINUX_BASE_IMAGES["opengl"]
-            self.prereqsTag = "opengl"
+            self.prereqsTag = "opengl-{ubuntu}"
+
+        self.baseImage = self.baseImage.format(
+            cuda=self.args.cuda, ubuntu=self.args.basetag
+        )
+        self.prereqsTag = self.prereqsTag.format(
+            cuda=self.args.cuda, ubuntu=self.args.basetag
+        )
 
     def _processPackageVersion(self, package, version):
 
